import json
import os

from tktl.commands.login import LogInCommand, SetApiKeyCommand


def test_set_api_key_command(capsys):
    cmd = SetApiKeyCommand()
    cmd.execute(api_key=None)
    out, err = capsys.readouterr()
    assert "API Key cannot be empty.\n" == err

    cmd.execute(api_key="ABC")
    assert os.path.exists(os.path.expanduser("~/.config/tktl/config.json"))
    with open(os.path.expanduser("~/.config/tktl/config.json"), "r") as j:
        d = json.load(j)
        assert d["api-key"] == "ABC"


def test_login_command(capsys):
    cmd = LogInCommand()
    k = os.environ["TEST_USER_API_KEY"]
    assert cmd.execute(k) is True
    out, err = capsys.readouterr()
    assert out == f"Authentication successful for user: {os.environ['TEST_USER']}\n"
    assert cmd.execute(None) is False


def test_login_fail_command(capfd):
    cmd = LogInCommand()
    assert cmd.execute() is False
    out, err = capfd.readouterr()
    assert err == "Authentication failed: no key provided\n"

    assert cmd.execute(api_key="INVALID_UUID") is False
    out, err = capfd.readouterr()
    assert err == "Authentication failed: Key format is invalid\n"



from .registration.decorator import Tktl  # noqa: F401
from .registration.endpoints import (  # noqa: F401
    ArrowEndpoint,
    BasicEndpoint,
    Endpoint,
    ProfiledEndpoint,
    TypedEndpoint,
)



from typing import Callable, Dict, List, Optional, Union

from pyarrow import Schema  # type: ignore
from pydantic import BaseModel

from tktl.core.t import EndpointKinds


class EndpointInfoSchema(BaseModel):
    name: str
    kind: EndpointKinds
    has_rest_sample_data: bool = False
    has_arrow_sample_data: bool = False
    profiling_supported: bool = False
    profiling_columns: Optional[List[str]] = []
    explain_input_names: Optional[List[str]] = []
    explain_input_example: Optional[List[Dict]] = []
    input_names: List[str] = []
    output_names: Union[str, List[str]] = []


class EndpointDefinition(BaseModel):
    name: str
    kind: str
    func: Callable
    input_schema: Optional[Union[Schema, bool]]
    output_schema: Optional[Union[Schema, bool]]
    has_sample_data: bool

    class Config:
        arbitrary_types_allowed = True



from typing import Dict

from tktl.core.managers.base import BaseConfigManager
from tktl.core.strings import HeaderStrings


class AuthConfigManager(BaseConfigManager):

    LOGIN_HEADERS = {
        "Accept": HeaderStrings.APPLICATION_JSON,
        "Content-Type": HeaderStrings.APPLICATION_URLENCODED,
    }

    @classmethod
    def get_api_key(cls):
        cfg = cls.get_tktl_config()
        if cfg and "api-key" in cfg:
            return cfg["api-key"]
        else:
            return None

    @classmethod
    def set_api_key(cls, api_key: str) -> Dict:
        cfg = cls.get_tktl_config()
        if cfg:
            if cfg["api-key"] == api_key:
                return {"api-key": api_key}
            cfg["api-key"] = api_key
            return cls.set_tktl_config(cfg)
        else:
            cfg = {"api-key": api_key}
            return cls.set_tktl_config(cfg)



from typing import List

from tktl.commands import BaseDeploymentApiCommand
from tktl.core.clients.http_client import interpret_response
from tktl.core.config import settings
from tktl.core.schemas.repository import Repository


class GetRepositories(BaseDeploymentApiCommand):
    def execute(self) -> List[Repository]:
        response = self.client.get(f"{settings.API_V1_STR}/models")
        return interpret_response(response=response, model=Repository)



import warnings
from typing import Any, Iterator, Sequence, Tuple, Union

import pandas  # type: ignore
import pandas.api.types as ptypes  # type: ignore

from tktl.core.exceptions import exceptions
from tktl.core.loggers import LOG
from tktl.core.utils import check_and_get_value


def validate_outputs(preds) -> Union[bool, Iterator]:
    if not isinstance(preds, pandas.Series):
        try:
            preds = pandas.Series(preds, dtype=float)
        except Exception as e:
            LOG.warning(
                f"Could not cast predictions to series: {e}, profiling won't happen"
            )
            return False
    if not ptypes.is_numeric_dtype(pandas.Series(preds)):
        LOG.warning("Function return types are not numeric, profiling won't happen")
        return False
    return True


def validate_binary(preds):
    if not 0 <= min(preds):
        raise exceptions.ValidationException(
            "Function output cannot be negative for endpoint kind binary"
        )

    if not max(preds) <= 1:
        raise exceptions.ValidationException(
            "Function output cannot exceed 1 for endpoint kind binary"
        )
    return True


def validate_shapes(
    x_frame: pandas.DataFrame, y_series: pandas.Series, preds: Sequence
):
    return len(x_frame) == len(preds) == len(y_series)


def data_frame_convertible(inputs):
    if len(inputs) > 1e6:
        warnings.warn(
            f"inputs is very large (N={len(inputs)}). Please consider using a smaller reference dataset."
        )
    if isinstance(inputs, pandas.DataFrame):
        return True
    try:
        pandas.DataFrame(inputs)
    except Exception as e:
        LOG.warning(
            f"Could not convert inputs to pd.DataFrame: {repr(e)}, profiling won't happen"
        )
        return False
    return True


def series_convertible(y, type_cast=None):
    if isinstance(y, pandas.Series):
        return True
    try:
        y = pandas.Series(y)
        if type_cast:
            y = y.astype(type_cast)
        y.name = y.name or "Outcome"

    except Exception as e:
        LOG.warning(
            f"Could not convert y to pandas series: {repr(e)}, profiling won't happen"
        )
        return False
    return True


def input_pandas_representation(
    value: Any, reset_index: bool = False
) -> pandas.DataFrame:
    """
    Transform endpoint inputs to pandas dataframe

    Parameters
    ----------
    value : Any
        Inputs
    reset_index : bool
        Optionally reset index on output dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    value = check_and_get_value(value, first_batch=True)
    df = pandas.DataFrame(value)
    df.columns = [str(c) for c in df.columns]
    if reset_index:
        df = df.reset_index(drop=True)
    return df


def output_pandas_representation(
    value: Any, reset_index: bool = False, type_cast=None
) -> pandas.Series:
    """
    Transform endpoint inputs to pandas series

    Parameters
    ----------
    value : Any
        Inputs
    reset_index : bool
        Optionally reset index on output dataframe
    type_cast : type
        Type to which to cast series

    Returns
    -------
    series : pandas.Series
    """
    value = check_and_get_value(value, first_batch=True)
    if isinstance(value, pandas.Series):
        if reset_index:
            value = value.reset_index(drop=True)
        if not value.name:
            value.name = "Outcome"
        return value
    else:
        series = pandas.Series(value, name="Outcome")
        if type_cast:
            series = series.astype(type_cast)
        return series


def get_input_and_output_for_profiling(
    x: pandas.DataFrame, y: pandas.Series
) -> Tuple[pandas.DataFrame, pandas.Series]:
    """
    Prepares inputs and outputs to be ready for profiling

    Parameters
    ----------
    x : pandas.DataFrame
    y : pandas.Series

    Returns
    -------
    x, y: Tuple[pandas.DataFrame, pandas.Series]
        Inputs and outputs
    """
    not_missing = [i for i, v in enumerate(y) if not pandas.isna(v)]
    n_missing = len(y) - len(not_missing)
    if n_missing > 0:
        warnings.warn(f"y contains {n_missing} missing values that will be dropped")
        return x.iloc[not_missing], y.iloc[not_missing]
    else:
        return x, y



from datetime import datetime
from typing import Dict, List, Optional, Union

import certifi
from pydantic import UUID4, BaseModel, validator

from tktl.constants import NON_EXISTING_DEPLOYMENT_DEFAULT_STATUS, URL_UNAVAILABLE
from tktl.core.config import settings
from tktl.core.t import AccessKind, ComputeTypeT, InstanceSizeT, ServiceKindT
from tktl.core.utils import flatten


class TablePrintableBaseModelMixin:
    def table_repr(self, subset: List[str] = None) -> Dict:
        ...


class Endpoint(BaseModel):
    name: str
    kind: str
    deployment_id: Optional[UUID4]
    profiling_supported: bool
    has_rest_sample_data: Optional[bool] = True
    has_arrow_sample_data: Optional[bool] = True

    class Config:
        validate_assignment = True

    @validator("deployment_id")
    def set_deployment_id(cls, deployment_id: UUID4):
        return deployment_id

    def table_repr(self, subset=None):
        as_dict = self.dict()
        as_dict["NAME"] = as_dict.pop("name")
        as_dict["KIND"] = str(as_dict.pop("kind"))
        as_dict["PROFILING SUPPORTED"] = str(as_dict.pop("profiling_supported"))
        as_dict["DEPLOYMENT ID"] = str(as_dict.pop("deployment_id"))
        return as_dict


class UserRepository(BaseModel):
    id: UUID4
    full_name: str


class RepositorySecret(BaseModel):
    id: UUID4
    repository_id: UUID4
    secret_name: str


class RepositoryDeployment(BaseModel, TablePrintableBaseModelMixin):
    id: UUID4
    created_at: datetime
    status: Optional[str]
    public_docs_url: Optional[str]

    rest_instance_type: Optional[ComputeTypeT]
    rest_instance_size: Optional[InstanceSizeT]
    rest_replicas: Optional[int]
    max_rest_replicas: Optional[int]
    #
    arrow_instance_type: Optional[ComputeTypeT]
    arrow_instance_size: Optional[InstanceSizeT]
    arrow_replicas: Optional[int]

    git_ref: str
    commit_hash: str
    n_endpoints: Optional[int]

    @validator("status", always=True)
    def validate_status(cls, value):
        return value if value else NON_EXISTING_DEPLOYMENT_DEFAULT_STATUS

    @validator("n_endpoints", always=True)
    def validate_n_endpoints(cls, value):
        return value or 0

    def table_repr(self, subset=None):
        as_dict = self.dict(exclude={"service_type", "endpoints",})
        as_dict["ID"] = str(as_dict.pop("id"))
        as_dict[
            "BRANCH @ COMMIT"
        ] = f"{as_dict.pop('git_ref')} @ {as_dict.pop('commit_hash')[0:7]}"
        as_dict["STATUS"] = as_dict.pop("status")
        as_dict["CREATED AT"] = str(as_dict.pop("created_at"))
        as_dict["REST DOCS URL"] = _format_http_url(as_dict.pop("public_docs_url"))
        as_dict["REST INSTANCE"] = self.get_instance(ServiceKindT.REST.value, as_dict)
        as_dict["ARROW INSTANCE"] = self.get_instance(ServiceKindT.ARROW.value, as_dict)
        as_dict["ENDPOINTS"] = as_dict.pop("n_endpoints")
        as_dict[
            "REST REPLICAS"
        ] = f"{as_dict.pop('max_rest_replicas')} ({as_dict.pop('rest_replicas')})"
        as_dict["ARROW REPLICAS"] = as_dict.pop("arrow_replicas")
        if subset:
            return {k: v for k, v in as_dict.items() if k in subset}
        return as_dict

    @staticmethod
    def get_instance(kind: ServiceKindT, values: Dict):
        if values[f"{kind}_instance_type"] is None:
            return ""
        else:
            return f"{values.pop(f'{kind}_instance_type').value}.{values.pop(f'{kind}_instance_size').value}"


class Repository(BaseModel, TablePrintableBaseModelMixin):
    id: UUID4
    ref_id: int
    repository_name: str
    repository_owner: str
    repository_description: Optional[str] = None
    access: AccessKind
    deployments: List[RepositoryDeployment]
    n_deployments: Optional[int] = None

    @validator("n_deployments", always=True)
    def validate_n_deployments(cls, _, values):
        return len(values["deployments"])

    def __hash__(self):
        return self.id.__hash__()

    def table_repr(self, subset=None):
        as_dict = self.dict(exclude={"ref_id", "deployments"})
        as_dict["ID"] = f"{as_dict.pop('id')}"
        as_dict[
            "FULL NAME"
        ] = f"{as_dict.pop('repository_owner')}/{as_dict.pop('repository_name')}"
        as_dict["DEPLOYMENTS"] = as_dict.pop("n_deployments")
        as_dict["ACCESS"] = f"{as_dict.pop('access').value}"
        desc = as_dict.pop("repository_description")
        as_dict["DESCRIPTION"] = f"{desc[0:20] + '...' if desc else '-'}"
        if subset:
            return {k: v for k, v in as_dict.items() if k in subset}
        return as_dict


class RepositoryList(BaseModel):
    __root__: List[Repository]

    def get_repositories(self) -> List[Repository]:
        return [r for r in self.__root__]

    def get_deployments(self) -> List[RepositoryDeployment]:
        return flatten([[d for d in r.deployments] for r in self.__root__])


class ReportResponse(BaseModel):
    deployment_id: UUID4
    endpoint_name: str
    report_type: str
    chart_name: Optional[str] = None
    variable_name: Optional[str] = None
    value: Union[List, Dict]


def _format_http_url(url: Optional[str], docs: bool = True) -> str:
    if url == URL_UNAVAILABLE or url is None:
        return URL_UNAVAILABLE
    if settings.LOCAL_STACK:
        return f"http://{url}:8000/{'docs' if docs else ''}"
    if settings.E2E:
        return f"http://{url}/{'docs' if docs else ''}".replace(
            ".local.taktile.com", "-rest.default.svc.cluster.local", 1
        )

    return f"https://{url}/{'docs' if docs else ''}"


def _format_grpc_url(url: Optional[str]) -> str:
    if url == URL_UNAVAILABLE or url is None:
        return URL_UNAVAILABLE
    if settings.LOCAL_STACK:
        return (
            f"grpc+tcp://{url}:5005"
            if (url and url != "UNAVAILABLE")
            else "UNAVAILABLE"
        )
    if settings.E2E:
        return (
            f"grpc+tcp://{url}:5005".replace(
                ".local.taktile.com", "-grpc.default.svc.cluster.local", 1
            )
            if (url and url != "UNAVAILABLE")
            else "UNAVAILABLE"
        )
    return f"grpc+tls://{url}:5005" if (url and url != "UNAVAILABLE") else "UNAVAILABLE"


def load_certs():
    with open(certifi.where(), "r") as cert:
        return cert.read()


Resources = Union[Repository, RepositoryDeployment, Endpoint, UserRepository]



import abc
import functools
import itertools
import os
import time
from typing import Generator, Generic, List, Sequence, TypeVar

import pandas  # type: ignore
import pyarrow  # type: ignore
import pyarrow.parquet as pq  # type: ignore
from cached_property import cached_property  # type: ignore

from tktl.core.config import settings
from tktl.core.exceptions import WrongPathError


def concatenate_urls(fst_part, snd_part):
    fst_part = fst_part if not fst_part.endswith("/") else fst_part[:-1]
    template = "{}{}" if snd_part.startswith("/") else "{}/{}"
    concatenated = template.format(fst_part, snd_part)
    return concatenated


class PathParser(object):
    LOCAL_DIR = 0
    LOCAL_FILE = 1
    GIT_URL = 2
    S3_URL = 3

    @classmethod
    def parse_path(cls, path):
        if cls.is_local_dir(path):
            return cls.LOCAL_DIR

        if cls.is_local_zip_file(path):
            return cls.LOCAL_FILE

        if cls.is_git_url(path):
            return cls.GIT_URL

        if cls.is_s3_url(path):
            return cls.S3_URL

        raise WrongPathError("Given path is neither local path, nor valid URL")

    @staticmethod
    def is_local_dir(path):
        return os.path.exists(path) and os.path.isdir(path)

    @staticmethod
    def is_local_zip_file(path):
        return os.path.exists(path) and os.path.isfile(path) and path.endswith(".zip")

    @staticmethod
    def is_git_url(path):
        return (
            not os.path.exists(path)
            and path.endswith(".git")
            or path.lower().startswith("git:")
        )

    @staticmethod
    def is_s3_url(path):
        return not os.path.exists(path) and path.lower().startswith("s3:")


def lru_cache(timeout: int, maxsize: int = 128, typed: bool = False):
    def wrapper_cache(func):
        func = functools.lru_cache(maxsize=maxsize, typed=typed)(func)
        func.delta = timeout * 10 ** 9
        func.expiration = time.monotonic_ns() + func.delta

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if time.monotonic_ns() >= func.expiration:
                func.cache_clear()
                func.expiration = time.monotonic_ns() + func.delta
            return func(*args, **kwargs)

        wrapped_func.cache_info = func.cache_info
        wrapped_func.cache_clear = func.cache_clear
        return wrapped_func

    return wrapper_cache


def flatten(x: Sequence) -> List:
    return list(itertools.chain.from_iterable(x))


LOADED_TYPE = TypeVar("LOADED_TYPE")


class DelayedLoader(Generic[LOADED_TYPE], metaclass=abc.ABCMeta):
    def __init__(self, path: str, columns: List[str]):
        number_of_columns = len(pq.read_table(path, memory_map=True).column_names)
        self._columns = columns
        self._path = path
        self._batch_size = settings.PARQUET_BATCH_DEFAULT_ROWS_READ // number_of_columns

    def load(self, first_batch: bool = False) -> LOADED_TYPE:
        if not first_batch:
            return self._load_full()
        else:
            return self._load_first_batch()

    def to_pandas_batches(self,) -> Generator[LOADED_TYPE, None, None]:
        return (self._post_process(x.to_pandas()) for x in self.to_batches())

    def to_batches(self) -> Generator[pyarrow.RecordBatch, None, None]:
        return pq.ParquetFile(self._path, memory_map=True).iter_batches(
            batch_size=self._batch_size, columns=self._columns
        )

    def _load_full(self) -> LOADED_TYPE:
        return self._post_process(
            pq.read_table(
                self._path, columns=self._columns, memory_map=True
            ).to_pandas()
        )

    def _load_first_batch(self) -> pandas.DataFrame:
        return next(self.to_pandas_batches())

    def _post_process(self, df: pandas.DataFrame) -> LOADED_TYPE:
        pass

    @cached_property
    def schema(self):
        return next(self.to_batches()).schema


class DelayedLoaderX(DelayedLoader[pandas.DataFrame]):
    def __init__(self, path: str, label: str, columns: List[str] = None):

        if not columns:
            columns = [
                c
                for c in pq.read_table(path, memory_map=True).column_names
                if not c.startswith("__index_level_") and c != label
            ]

        super().__init__(path=path, columns=columns)

    def _post_process(self, df: pandas.DataFrame) -> pandas.DataFrame:
        return df


class DelayedLoaderXProfile(DelayedLoaderX):
    def _post_process(self, df: pandas.DataFrame) -> pandas.DataFrame:
        df.columns = [str(c) for c in df.columns]
        return df


class DelayedLoaderY(DelayedLoader[pandas.Series]):
    def __init__(self, path: str, label: str):
        self._label = label
        super().__init__(path=path, columns=[label])

    def _post_process(self, df: pandas.DataFrame) -> pandas.Series:
        return df[self._label]


def check_and_get_value(value, first_batch=False):
    if isinstance(value, DelayedLoader):
        return value.load(first_batch=first_batch)
    else:
        return value



import random
from typing import Any, Callable, List

import numpy as np
import pandas
import pandas as pd
import pytest
from pydantic import BaseModel

from tktl import Tktl
from tktl.core.exceptions.exceptions import ValidationException
from tktl.core.utils import DelayedLoaderX, DelayedLoaderY
from tktl.registration.validation import (
    get_input_and_output_for_profiling,
    input_pandas_representation,
    output_pandas_representation,
)


@pytest.mark.parametrize(
    "X,y",
    [
        (pd.DataFrame({"a": [1, 2, 3]}), pd.Series([4, 5, 5])),  # pandas
        (np.array([[1], [2], [3]]), np.array([4, 5, 5])),  # numpy
        ({"a": [1, 2, 3]}, [4, 5, 5]),  # base
    ],
)
def test_creation(X, y):
    tktl = Tktl()

    @tktl.endpoint(X=X, y=y, kind="tabular")
    def predict(X):
        return [0] * 3

    endpoint = tktl.endpoints[0]
    assert endpoint.input_schema.pandas_convertible is True
    assert endpoint.output_schema.pandas_convertible is True
    pred = endpoint.func(pd.DataFrame(X))
    assert pred == [0, 0, 0]


@pytest.mark.parametrize("kind", ["tabular", "regression", "binary"])
def test_tabular_kinds(kind):
    tktl = Tktl()
    X = pd.DataFrame({"a": [1, 2, 3]})
    y = pd.Series([4, 5, 5])

    @tktl.endpoint(X=X, y=y, kind=kind)
    def predict(X):
        return [0] * len(X)

    endpoint = tktl.endpoints[0]
    assert isinstance(endpoint.input_schema.value, pd.DataFrame)
    assert isinstance(endpoint.output_schema.value, pd.Series)
    pred = endpoint.func(pd.DataFrame(X))
    assert pred == [0, 0, 0]


def test_custom_kind():
    tktl = Tktl()

    @tktl.endpoint(kind="custom")
    def predict_untyped(x):
        return x

    @tktl.endpoint(kind="custom", payload_model=Any, response_model=Any)
    def predict_untyped_any(x):
        return x

    class Payload(BaseModel):
        a: List

    class Response(BaseModel):
        a: int
        b: float
        c: List

    @tktl.endpoint(kind="custom", payload_model=Payload, response_model=Response)
    def predict_typed_input_schemas(x):
        return x

    for endpoint in tktl.endpoints:
        assert isinstance(endpoint.func, Callable)
        assert endpoint.func(["foo"]) == ["foo"]

    x = pd.DataFrame([{"a": 1, "b": 2}, {"a": 1, "b": 2}, {"a": 1, "b": 2}])
    y = pd.Series([1, 2, 3])

    @tktl.endpoint(kind="custom", X=x, y=y)
    def predict_typed_input_frames(x):
        return x

    schema = tktl.endpoints[3].input_schema.pydantic.schema()
    assert schema["title"] == "DataFrame__predict_typed_input_frames__input"
    assert schema["example"] == [{"a": 1, "b": 2}]


def test_func_shape():
    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = [4, 5, 5]

    @tktl.endpoint(X=X, y=y)
    def predict(X):
        return [0] * (len(X) - 1)

    assert tktl.endpoints[0].profiling_supported is False


def test_input_shape():
    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = [4, 5]

    @tktl.endpoint(X=X, y=y)
    def predict(X):
        return [0] * len(X)

    assert tktl.endpoints[0].profiling_supported is False


def test_validate_func():
    tktl = Tktl()

    X = {"a": [1, 2, 3]}
    y = [4, 5, 5]

    @tktl.endpoint(X=X, y=y, kind="regression")
    def predict_regression(X):
        return ["a", "b", "c"]  # must be numeric

    assert tktl.endpoints[0].profiling_supported is False

    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = [4, 5, 5]

    with pytest.raises(ValidationException):

        @tktl.endpoint(X=X, y=y, kind="binary")
        def predict_binary(X):
            return [-2, 0, 1.5]  # must be in [0, 1]

        _ = tktl.endpoints[0].profiling_supported


def test_unknown_kind():
    with pytest.raises(ValidationException):
        tktl = Tktl()
        X = {"a": [1, 2, 3]}
        y = [4, 5, 5]

        @tktl.endpoint(X=X, y=y, kind="unknown kind")
        def predict(X):
            return [0] * len(X)


def test_missing():
    tktl = Tktl()
    X = pd.DataFrame({"a": [1, 2, 3]})
    y = pd.Series([0.1, 0.2, None])

    @tktl.endpoint(X=X, y=y, kind="regression")
    def predict(X):
        return [0] * len(X)

    with pytest.warns(UserWarning):
        x_raw = input_pandas_representation(
            value=tktl.endpoints[0].input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=tktl.endpoints[0].output_schema.value, reset_index=False
        )
        get_input_and_output_for_profiling(x=x_raw, y=y_raw)


def test_large():
    tktl = Tktl()
    n = int(1e6) + 1
    X = pd.DataFrame({"a": np.random.uniform(size=n)})
    y = pd.Series(np.random.uniform(size=n))

    with pytest.warns(UserWarning):

        @tktl.endpoint(X=X, y=y, kind="regression")
        def predict(X):
            return [0] * len(X)


def test_custom():
    tktl = Tktl()

    class Payload(BaseModel):
        a: str

    class Response(BaseModel):
        a: str

    @tktl.endpoint(kind="custom", payload_model=Payload, response_model=Response)
    def string_endpoint(x):
        return x


def test_delayed_loaded_frame(create_proj):
    tktl = Tktl()
    path = "sample_project/assets/loans_test.pqt"

    @tktl.endpoint(X=path, y="Repaid", kind="binary")
    def repayment(df):
        return pandas.Series([random.random() for _ in range(len(df))])

    assert tktl.endpoints[0].profiling_supported is True
    input_schema = tktl.endpoints[0].input_schema
    output_schema = tktl.endpoints[0].output_schema
    assert isinstance(input_schema.value, DelayedLoaderX)
    assert isinstance(output_schema.value, DelayedLoaderY)



from click.testing import CliRunner

from tktl import main


def test_health(logged_in_context, test_user_deployed_repos):
    """Test the CLI."""
    runner = CliRunner()
    for repo, branch, endpoint in test_user_deployed_repos[2:]:
        result = runner.invoke(main.health, ["-r", repo, "-b", branch, "-s", "grpc"])
        assert result.exit_code == 0
        result = runner.invoke(main.health, ["-r", repo, "-b", branch, "-s", "rest"])
        assert result.exit_code == 0


def test_health_fail(test_user_deployed_repos):
    repo, branch, endpoint = test_user_deployed_repos[0]
    runner = CliRunner()
    result = runner.invoke(main.health, ["-r", repo, "-b", branch, "-s", "rest"])
    # User not logged in
    assert result.exit_code == 1


def test_health_local(logged_in_context):
    runner = CliRunner()
    result = runner.invoke(main.health, ["-s", "rest"])
    assert result.exc_info[0] == SystemExit

    result = runner.invoke(main.health, ["-s", "rest", "-l"])
    assert result.exc_info[0] == SystemExit

    # No running local service
    assert result.exit_code == 1



class CLIStrings:
    INVALID_INPUT = "Invalid Input: "
    AUTH_ERROR_MISSING_KEY = "No keys found locally. Please login first."
    REQUEST_FAILED = "Request Failed: "


class HeaderStrings:
    APPLICATION_JSON = "application/json"
    APPLICATION_URLENCODED = "application/x-www-form-urlencoded"
